#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <inttypes.h>
#include <math.h>
#include <string.h>

//#define DUMP_DELTAS
//#define BIG_ENDIAN
//#define REPORT_ODDNESS

static void writeint32LE(uint8_t *c, uint32_t val)
{
    c[0] = val & 0xFF;
    val >>= 8;
    c[1] = val & 0xFF;
    val >>= 8;
    c[2] = val & 0xFF;
    val >>= 8;
    c[3] = val & 0xFF;
}

static void writeint16LE(uint8_t *c, uint32_t val)
{
    c[0] = val & 0xFF;
    val >>= 8;
    c[1] = val & 0xFF;
}

#ifdef BIG_ENDIAN
static uint16_t readint16(uint8_t *c)
{
    uint16_t total = c[0];
    total *= 0x100;
    total += c[1];
    return total;
}

static uint32_t readint32(uint8_t *c)
{
    uint32_t total = c[0];
    total *= 0x100;
    total += c[1];
    total *= 0x10000;
    total += readint16(c+2);
    return total;
}
#else
static uint16_t readint16(uint8_t *c)
{
    uint16_t total = c[1];
    total *= 0x100;
    total += c[0];
    return total;
}

static uint32_t readint32(uint8_t *c)
{
    uint32_t total = c[3];
    total *= 0x100;
    total += c[2];
    total *= 0x10000;
    total += readint16(c);
    return total;
}
#endif

#ifdef DUMP_DELTAS
//#define TOTAL_FRAMES (41620*2)
#define TOTAL_FRAMES (1<<22)
int frame_diff[TOTAL_FRAMES];
int frame_deltas[TOTAL_FRAMES][32];
int saved_frames = 0;
#endif

struct decode_context_s;

int do_file(FILE *infile, FILE *outfile, int *got_oddness);
int decode_block(struct decode_context_s *dc);

int main(int argc, char **argv)
{
    FILE *infile = fopen(argv[1], "rb");
    uint8_t buf[0x10];

    unsigned long cur_off;
    if (!infile) return 1;

#ifdef DUMP_DELTAS
    FILE *dumpfile = fopen("delta.dmp", "wb");
#endif

    /* read into and out of "files" */
    for (int file_number = 0; 0x10 == fread(buf, 1, 0x10, infile); file_number ++)
    {
        int got_oddness = 0;

        cur_off = ftell(infile) - 0x10;

        if (readint32(&buf[0]) != 0x10 || readint32(&buf[4]) != 0x10 || readint32(&buf[8]) != 0 || readint16(&buf[12]) != 1 || readint16(&buf[14]) != 0x11)
        {
            fprintf(stderr, "didn't see file intro at 0x%lx\n", cur_off);
            return 1;
        }

        {
            char nameo[] = "BGMXX.wav";
            snprintf(nameo, sizeof(nameo), "BGM%02d.wav", file_number+1);

            FILE *outfile = fopen(nameo, "wb");
            if (!outfile)
            {
                fprintf(stderr, "failed to open %s\n", nameo);
                return 1;
            }

            printf("%s: file %d starting at %08lx\n", nameo, file_number, cur_off);

            // the goodness!
            if (do_file(infile, outfile, &got_oddness) != 0)
            {
                fprintf(stderr, "file failed\n");
                fclose(outfile);
                return 1;
            }

            if (EOF == fclose(outfile))
            {
                perror("writing error");
                return 1;
            }
        }

        cur_off = ftell(infile);

        if (0x10 != fread(buf, 1, 0x10, infile))
        {
            perror("reading file outro");
            return 1;
        }

        if (!got_oddness)
        {
            // the file processor didn't get the 2nd oddness, get it now

            if (readint32(&buf[0]) != 0xf0 || readint32(&buf[4]) != 0x10 || readint32(&buf[12]) != 0)
            {
                fprintf(stderr, "didn't see file outro at 0x%lx\n", cur_off);
                return 1;
            }

#ifdef REPORT_ODDNESS
            printf("file end oddness 2 %08lx\n", (unsigned long)readint32(&buf[8]));
#endif
        }

        {
            unsigned long end_offset = ftell(infile);
            unsigned long padded_offset = (end_offset+0x7ff)/0x800*0x800;
            printf("end at %08lx, pad to %08lx\n", end_offset, padded_offset);
            fseek(infile, padded_offset - end_offset, SEEK_CUR);
        }

        printf("\n");
    }

#ifdef DUMP_DELTAS
        {
            int steps = 32;
            fwrite(&saved_frames, sizeof(saved_frames), 1, dumpfile);
            fwrite(&steps, sizeof(steps), 1, dumpfile);

            for (int i = 0; i < saved_frames; i++)
            {
                double d = frame_diff[i];
                fwrite(&d, sizeof(d), 1, dumpfile);
            }

            for (int i = 0; i < 32; i++)
            {
                for (int j = 0; j < saved_frames; j++)
                {
                    double d = frame_deltas[j][i];
                    fwrite(&d, sizeof(d), 1, dumpfile);
                }
            }
        }
#endif

    printf("done!\n");

    return 0;

}

struct decode_context_s
{
    uint32_t seq;   // seemingly 24 bits
    int streams;
    int channels;
    int steps;
    long frame_size;
    long frame_samples;
    long subblock_size;
    uint8_t *frame_buf;
    int16_t *samples;

    int *step_idx; // [stream][channel]
    long *hist;    // [stream][channel]

    int subblocks;
    long total_frames;
    long total_subblocks;

    int clipcount;
    long errmag;

#ifdef DUMP_DELTAS
    int *prev_frame_init;  // [stream][channel]
    int *total_delta; // [stream][channel][step index]
#endif

    FILE *infile;
    FILE *outfile;
};

/* make a header for PCM .wav */
/* buffer must be 0x2c bytes */
void make_wav_header(uint8_t * buf, int32_t sample_count, int32_t sample_rate, int channels) {
    size_t bytecount;

    bytecount = sample_count*channels*sizeof(int16_t);

    /* RIFF header */
    memcpy(buf+0, "RIFF", 4);
    /* size of RIFF */
    writeint32LE(buf+4, (int32_t)(bytecount+0x2c-8));

    /* WAVE header */
    memcpy(buf+8, "WAVE", 4);

    /* WAVE fmt chunk */
    memcpy(buf+0xc, "fmt ", 4);
    /* size of WAVE fmt chunk */
    writeint32LE(buf+0x10, 0x10);

    /* compression code 1=PCM */
    writeint16LE(buf+0x14, 1);

    /* channel count */
    writeint16LE(buf+0x16, channels);

    /* sample rate */
    writeint32LE(buf+0x18, sample_rate);

    /* bytes per second */
    writeint32LE(buf+0x1c, sample_rate*channels*sizeof(int16_t));

    /* block align */
    writeint16LE(buf+0x20, (int16_t)(channels*sizeof(int16_t)));

    /* significant bits per sample */
    writeint16LE(buf+0x22, sizeof(int16_t)*8);

    /* PCM has no extra format bytes, so we don't even need to specify a count */

    /* WAVE data chunk */
    memcpy(buf+0x24, "data", 4);
    /* size of WAVE data chunk */
    writeint32LE(buf+0x28, (int32_t)bytecount);
} 

int do_file(FILE *infile, FILE *outfile, int *got_oddness)
{
    int rc = 1;
    unsigned long cur_off;

    uint8_t header_buf[0x10];

    struct decode_context_s dc;
    memset(&dc, 0, sizeof(dc));


    /* loop through blocks */
    for (long block_number = 0; ; block_number ++)
    {
        uint32_t block_size;

        cur_off = ftell(infile);

        if (0x10 != fread(header_buf, 1, 0x10, infile))
        {
            fprintf(stderr, "error reading block header at 0x%lx\n", cur_off);
            goto cleanup;
        }

        if (1 != readint16(&header_buf[0]))
        {
            if (0xf0 == readint32(&header_buf[0]))
            {
                if (0x10 != readint32(&header_buf[4]) || 0 != readint32(&header_buf[8]))
                {

                    if (0x10 != readint32(&header_buf[4]) || 0 != readint32(&header_buf[12]))
                    {
                        fprintf(stderr, "wrong stuff for file outro at %08lx\n", cur_off);
                        goto cleanup;
                    }
                    else
                    {
                        *got_oddness = 1;
#ifdef REPORT_ODDNESS
                        printf("file end oddness 2 (no 1) %08lx\n", (unsigned long)readint32(&header_buf[8]));
#endif
                        // done in an even more unusual way (file 21)
                        break;
                    }
                }
#ifdef REPORT_ODDNESS
                printf("file end oddness 1 %08lx\n", (unsigned long)readint32(&header_buf[12]));
#endif
                // done!
                break;
            }

            fprintf(stderr, "expected 1 at start of block header at 0x%lx\n", cur_off);

            goto cleanup;
        }

        if (0x11 != readint16(&header_buf[2]))
        {
            fprintf(stderr, "expected 0x11 at 0x%lx\n", cur_off);
            goto cleanup;
        }

        block_size = readint32(&header_buf[4]);

        if (0 != readint32(&header_buf[8]))
        {
            fprintf(stderr, "expected 0 word in block header at 0x%lx\n", cur_off);
            goto cleanup;
        }

        if (block_number == 0)
        {
            // ****** header block

            memset(&dc, 0, sizeof(dc));

            unsigned long mtaf_header_start = ftell(infile) - 0x10;

            uint8_t mtaf_header_buf[0xB0];

            if (readint32(&header_buf[12]) != 0)
            {
                fprintf(stderr, "expected 0 frames in header at %08lx\n", cur_off);
                goto cleanup;
            }

            // master MTAF header (mostly useless)
            cur_off = ftell(infile);
            if (0x40 != fread(mtaf_header_buf, 1, 0x40, infile))
            {
                fprintf(stderr, "error reading MTAF header at 0x%08lx\n", cur_off);
                goto cleanup;
            }

            if (memcmp(&mtaf_header_buf[0], "MTAF", 4))
            {
                fprintf(stderr, "no MTAF header at 0x%08lx\n", cur_off);
                goto cleanup;
            }

            //const uint32_t pseudo_size = readint32(&mtaf_header_buf[4]);

            // check the rest is clear
            for (int i = 8; i < 0x20; i++)
            {
                if (mtaf_header_buf[i] != 0)
                {
                    fprintf(stderr, "unexpected nonzero in MTAF header at 0x%08lx\n", cur_off+i);
                    goto cleanup;
                }
            }

            // ignore the rest for now

            // HEAD chunk header
            cur_off = ftell(infile);
            if (8 != fread(mtaf_header_buf, 1, 8, infile))
            {
                fprintf(stderr, "error reading HEAD chunk at 0x%08lx\n", cur_off);
                goto cleanup;
            }

            if (memcmp(&mtaf_header_buf[0], "HEAD", 4))
            {
                fprintf(stderr, "no HEAD chunk at 0x%08lx\n", cur_off);
                goto cleanup;
            }

            uint32_t mtaf_head_chunk_size = readint32(&mtaf_header_buf[4]);
            if (mtaf_head_chunk_size != 0xB0)
            {
                fprintf(stderr, "unexpected size for MTAF header at 0x%08lx\n", cur_off);
                goto cleanup;
            }

            cur_off = ftell(infile);
            if (0xB0 != fread(mtaf_header_buf, 1, 0xB0, infile))
            {
                fprintf(stderr, "error reading HEAD chunk body at 0x%08lx\n", cur_off);
                goto cleanup;
            }

            // seems to always be the case
            dc.channels = 2;
            dc.streams = mtaf_header_buf[0x19];

            // just for kicks, get loop and sample info
            uint32_t loop_start_samples = readint32(&mtaf_header_buf[0x10]);
            uint32_t total_samples = readint32(&mtaf_header_buf[0x14]);

            // assume the rest of the format specifics
            dc.seq = 1;
            dc.steps = 32;
            dc.frame_size = 0x80;
            dc.frame_samples = dc.frame_size * 2;
            dc.subblock_size = dc.frame_size * dc.channels + 0x10;
            dc.frame_buf = calloc(dc.frame_size, sizeof(uint8_t));
            dc.samples = calloc(dc.frame_size, sizeof(int16_t) * dc.frame_samples * dc.streams * dc.channels);

#ifdef DUMP_DELTAS
            dc.prev_frame_init = calloc(dc.streams*dc.channels, sizeof(int));
            dc.total_delta = calloc(dc.streams*dc.channels*dc.steps, sizeof(int));
#endif
            dc.step_idx = calloc(dc.streams*dc.channels, sizeof(int));
            dc.hist = calloc(dc.streams*dc.channels, sizeof(long));

            dc.infile = infile;
            dc.outfile = outfile;

            // check some standard stuff
            if (   0 != readint32(&mtaf_header_buf[0]) ||
                    0x7F != readint32(&mtaf_header_buf[8]) ||
                    0x40 != readint32(&mtaf_header_buf[0xC]) ||
                    0 != readint16(&mtaf_header_buf[0x1a]) ||
                    0 != readint32(&mtaf_header_buf[0x24])) // ||
                //5 != readint32(&mtaf_header_buf[0x28])) ||
                //(dc.streams==3 ? 12:0) != readint32(&mtaf_header_buf[0x2c]))
            {
                fprintf(stderr, "unexpected header values at 0x%08lx\n", cur_off);
                goto cleanup;
            }

            // 0 streams should be impossible
            if (dc.streams == 0)
            {
                fprintf(stderr, "0 streams at 0x%08lx\n", cur_off);
                goto cleanup;
            }

            // check the other stream count indicator
            if (dc.streams*0x10 != mtaf_header_buf[0x18])
            {
                fprintf(stderr, "secondary stream count mismatch at 0x%08lx\n", cur_off);
                goto cleanup;
            }
            // check total channel count
            if (dc.channels*dc.streams != readint32(&mtaf_header_buf[4]))
            {
                fprintf(stderr, "total channel count does not match stream count at 0x%08lx\n", cur_off);
                goto cleanup;
            }
            // check loop points as frame counts
            if (readint32(&mtaf_header_buf[0x1C]) != loop_start_samples/0x100 ||
                    readint32(&mtaf_header_buf[0x20]) != total_samples/0x100)
            {
                fprintf(stderr, "loop frame count mismatch at 0x%lx\n", cur_off);
                goto cleanup;
            }

            // check that rest is clear
            for (int i = 0x30; i < mtaf_head_chunk_size; i++)
            {
                if (mtaf_header_buf[i] != 0)
                {
                    fprintf(stderr, "unexpected nonzero in HEAD chunk at 0x%lx\n", cur_off+i);
                    goto cleanup;
                }
            }

            // check TRKP chunks
            for (int i = 0; i < 16; i++)
            {
                cur_off = ftell(infile);
                if (0x70 != fread(mtaf_header_buf, 1, 0x70, infile))
                {
                    fprintf(stderr, "error reading TRKP chunk #%d at 0x%lx\n", i, cur_off);
                    goto cleanup;
                }

                if (memcmp(&mtaf_header_buf[0], "TRKP\x68\0\0", 8))
                {
                    fprintf(stderr, "missing or unusual TRKP chunk #%d at 0x%lx\n", i, cur_off);
                    goto cleanup;
                }
            }

            // check for grand finale, DATA
            cur_off = ftell(infile);

            if (8 != fread(mtaf_header_buf, 1, 8, infile))
            {
                fprintf(stderr, "error reading DATA chunk header at 0x%lx\n", cur_off);
                goto cleanup;
            }

            if (memcmp(&mtaf_header_buf[0], "DATA", 4))
            {
                fprintf(stderr, "missing DATA header at 0x%lx\n", cur_off);
                goto cleanup;
            }

            //const uint32_t pseudo_data_size = readint32(&mtaf_header_buf[4]);

            // we should have now exhausted the header block
            cur_off = ftell(infile);
            if (mtaf_header_start+block_size != cur_off)
            {
                fprintf(stderr, "did not read exactly header block, started at 0x%lx and ended at 0x%lx (expected 0x%lx)\n",
                        mtaf_header_start, cur_off, mtaf_header_start+block_size);
                goto cleanup;
            }


            // write the WAVE header
            {
                uint8_t wave_header[0x2c];
                make_wav_header(wave_header, total_samples, 44100, dc.streams * dc.channels);
                if (0x2c != fwrite(wave_header, 1, 0x2c, dc.outfile))
                {
                    perror("error writing wav header");
                    goto cleanup;
                }
            }
        }
        else /* block is not header */
        {
            // ****** audio (or padding) block

            uint32_t subblocks = readint32(&header_buf[12]) * dc.streams;

            if (subblocks == 0)
            {
                // skip padding
                fseek(infile, block_size-0x10, SEEK_CUR);
                cur_off = (unsigned long)ftell(infile);
                continue;
            }
            else if (subblocks * dc.subblock_size + 0x10 != block_size)
            {
                fprintf(stderr, "block size mismatch at 0x%lx, %lx != %lx (probably %d streams)\n", cur_off, (unsigned long)subblocks * dc.subblock_size + 0x10, (unsigned long)block_size,
                        (block_size-0x10)/readint32(&header_buf[12])/0x110);
                goto cleanup;
            }

            dc.subblocks = subblocks;

            if (decode_block(&dc) != 0)
            {
                fprintf(stderr, "decode failed\n");
                goto cleanup;
            }
        } // end if past header
    } // end block loop

    printf("%lf error per frame (%ld/%ld)\n", (double)dc.errmag/dc.total_frames, dc.errmag, dc.total_frames);

    // success!
    rc = 0;

cleanup:
    if (dc.frame_buf)
    {
        free(dc.frame_buf);
        dc.frame_buf = NULL;
    }
    if (dc.samples)
    {
        free(dc.samples);
        dc.samples = NULL;
    }
#ifdef DUMP_DELTAS
    if (dc.prev_frame_init)
    {
        free(dc.prev_frame_init);
        dc.prev_frame_init = NULL;
    }
    if (dc.total_delta)
    {
        free(dc.total_delta);
        dc.total_delta = NULL;
    }
#endif
    if (dc.step_idx)
    {
        free(dc.step_idx);
        dc.step_idx = NULL;
    }
    if (dc.hist)
    {
        free(dc.hist);
        dc.hist = NULL;
    }

    return rc;
}

int index_table[16] = {
    -1, -1, -1, -1, 2, 4, 6, 8,
    -1, -1, -1, -1, 2, 4, 6, 8
}; 

// .8 fixed point
int step_table[32] = {
    //458,559,650,779,940,1132,1352,1614,1938,2298,2736,3277,3880,4652,5554,6606,7854,9357,11153,13311,15870,18862,22477,26797,31916,38032,45310,53932,64268,76554,91188,108642
    455,558,651,779,940,1131,1354,1613,1937,2300,2734,3277,3882,4652,5553,6607,7855,9357,11153,13312,15871,18862,22477,26796,31915,38030,45309,53931,64264,76554,91181,108629,
};

int delta_table[16] = {
    1,  3,  5,  7,  9, 11, 13, 15,
    -1, -3, -5, -7, -9,-11,-13,-15
};

// convert from .8 fixed to nearest int
int16_t roundy(long v)
{
    if (v > 0)
    {
        return (v+0x80)/0x100;
    }
    else
    {
        return -(((-v)+0x80)/0x100);
    }
}

int decode_block(struct decode_context_s *dc)
{
    uint8_t head_buf[0x10];

    /* loop through subblocks */
    for (int subblock_number = 0; subblock_number < dc->subblocks; subblock_number++, dc->total_subblocks++)
    {
        unsigned long cur_off = ftell(dc->infile);
        uint8_t stream;

        if (0x10 != fread(head_buf, 1, 0x10, dc->infile))
        {
            fprintf(stderr, "subblock header read error at 0x%lx\n", cur_off);
            return 1;
        }

        uint32_t subblock_id = readint32(&head_buf[0]);
        //printf("%02x%02x%02x%02x %08x\n", head_buf[0], head_buf[1], head_buf[2], head_buf[3], (unsigned int)subblock_id);
        stream = subblock_id & 0xFF;
        if (stream >= dc->streams)
        {
            fprintf(stderr, "bad stream id at 0x%lx\n", cur_off);
            return 1;
        }
        if (dc->seq != (subblock_id>>8))
        {
            fprintf(stderr, "bad seq at 0x%lx %u != %u\n", cur_off, dc->seq, (unsigned int)subblock_id>>8);
            return 1;
        }

        // one frame per channel
        for (int c = 0; c < dc->channels; c++, dc->total_frames ++)
        {
            int16_t init_idx = readint16(&head_buf[4+c*2]);
            int16_t init_hist = readint16(&head_buf[8+c*4]);
            int hist;
            int step_idx = dc->step_idx[stream * dc->channels + c];

            if (init_idx < 0 || init_idx > 31)
            {
                fprintf(stderr, "step idx out of range at 0x%lx ch %d\n", cur_off, c);
                return 1;
            }
            if (0 != readint16(&head_buf[10+c*4]))
            {
                fprintf(stderr, "exp. zero after hist at 0x%lx ch %d\n", cur_off, c);
                return 1;
            }

            //
#ifdef DUMP_DELTAS
            if (dc->total_subblocks >= dc->streams)
            {
                if (saved_frames < TOTAL_FRAMES)
                {
                    memcpy(frame_deltas[saved_frames],&dc->total_delta[(stream * dc->channels + c) * dc->steps], sizeof(frame_deltas[saved_frames]));
                    frame_diff[saved_frames] = (init_hist - dc->prev_frame_init[stream * dc->channels + c]);
                    saved_frames++;
                }
            }

            dc->prev_frame_init[stream * dc->channels + c] = init_hist;
#endif


            dc->errmag += abs(roundy(dc->hist[stream * dc->channels + c])-init_hist);
            hist = init_hist*0x100;

            if (step_idx != init_idx)
            {
                fprintf(stderr, "step_idx does not match at 0x%lx, %d!=%d\n",cur_off,step_idx, init_idx);
                return 1;
            }
            step_idx = init_idx;

#ifdef DUMP_DELTAS
            memset(&dc->total_delta[(stream * dc->channels + c) * dc->steps], 0, sizeof(int) * dc->steps);
#endif

            cur_off = ftell(dc->infile);

            if (dc->frame_size != fread(dc->frame_buf, 1, dc->frame_size, dc->infile))
            {
                fprintf(stderr, "error reading frame at %08lx\n", cur_off);
                return 1;
            }

            for (unsigned int j = 0; j < dc->frame_samples; j++)
            {
                uint8_t nibble;
                if (j%2!=1)
                {
                    // low nibble first
                    nibble = (dc->frame_buf[j/2])&0xf;
                }
                else
                {
                    // high nibble last
                    nibble = (dc->frame_buf[j/2])>>4;
                }

                hist += delta_table[nibble] * step_table[step_idx];

                if (hist > INT16_MAX*0x100L)
                {
                    hist = INT16_MAX*0x100L;
                    dc->clipcount++;
                }
                if (hist < INT16_MIN*0x100L)
                {
                    hist = INT16_MIN*0x100L;
                    dc->clipcount++;
                }

                dc->samples[(j * dc->streams + stream) * dc->channels + c] = roundy(hist);

#ifdef DUMP_DELTAS
                dc->total_delta[((stream*dc->channels+c)*dc->steps)+step_idx] += delta_table[nibble];
#endif
                step_idx += index_table[nibble];
                if (step_idx < 0)
                {
                    step_idx = 0;
                }
                if (step_idx > dc->steps-1)
                {
                    step_idx = dc->steps-1;
                }
            } /* end sample loop */

            // update state
            dc->step_idx[stream * dc->channels + c] = step_idx;
            dc->hist[stream * dc->channels + c] = hist;
        } /* end frame/channel loop */

        dc->seq ++;

        // assume subblocks for streams occur in order
        if (stream == dc->streams-1)
        {
            long samples_to_write = dc->frame_samples * dc->streams * dc->channels;
#ifndef DUMP_DELTAS
            for (int i = 0; i < samples_to_write; i++)
            {
                uint8_t buf[2];
                writeint16LE(buf, dc->samples[i]);
                if (1 != fwrite(buf, sizeof(int16_t), 1, dc->outfile))
                {
                    perror("writing output");
                    return 1;
                }
            }
#endif
        }
    } /* end subblock loop */

    //printf("clipped %d\n", clipcount);
    return 0;
}

